# Copyright 2007, Didier Bur
#
# Permission to use, copy, modify, and distribute this software for
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#----------------------------------------------------------------------------
# Name : Align Tool
# Description : A tool to create translate and rotate a selection of entities.
# Menu Item : Tools / Align
# Usage : 1. Make a selection first
#         2. Click a start origin, X direction, and Y direction
#         3. Click a target origin, X direction, and Y direction
# Date : 08 May 2007
# Type : Tool
# History: 10 May 2007: added a context menu added (align or align group/component)
#                       added the align from group/component axis
#                       supressed confirmation dialog box
#                       Mac problem solved (commit_operation), thanks to R. Wilson
#----------------------------------------------------------------------------

module Align

  class AlignTool

    def initialize
      @ip1 = nil
      @ip2 = nil
      @ip3 = nil
      @ip4 = nil
      @ip5 = nil
      @ip6 = nil
      @xdown = 0
      @ydown = 0
      $align_points=[]
      @sel = Sketchup.active_model.selection
      @state = 0

      if @sel.empty?
        UI.messagebox("No selection to align.")
        Sketchup.send_action "selectSelectionTool:"
      end

      if Align.component_selected and $comp == true
        @mro_axes = Sketchup.active_model.rendering_options["DisplayInstanceAxes"]
        Sketchup.active_model.rendering_options["DisplayInstanceAxes"] = true
        t = @sel[0].transformation
        $align_points[0]=t.origin
        $align_points[1]=t.origin.offset t.xaxis
        $align_points[2]=t.origin.offset t.yaxis
        @state = 3
      end

    end

    # When the tool is first selected.
    def activate
      Sketchup.active_model.start_operation "Align"
      @ip1 = Sketchup::InputPoint.new
      @ip2 = Sketchup::InputPoint.new
      @ip3 = Sketchup::InputPoint.new
      @ip4 = Sketchup::InputPoint.new
      @ip5 = Sketchup::InputPoint.new
      @ip6 = Sketchup::InputPoint.new
      @ip = Sketchup::InputPoint.new
      @drawn = false
      self.reset(nil)
    end

    # When the tool is deactivated because a different tool was selected
    def deactivate(view)
      view.invalidate if @drawn
    end

    # Whenever the user moves the mouse.
    def onMouseMove(flags, x, y, view)
      #if( @state == 0 )
      # We are getting the first point (start origin)
      @ip.pick view, x, y
      if( @ip != @ip1 )
        view.invalidate if( @ip.display? or @ip1.display? )
        @ip1.copy! @ip
        # set the tooltip that should be displayed to this point
        #view.tooltip = @ip1.tooltip
      end
      #end

      case @state
      when 0
        view.tooltip = "Click start origin point"
      when 1
        view.tooltip = "Click start X axis orientation"
      when 2
        view.tooltip = "Click start Y axis orientation"
      when 3
        view.tooltip = "Click end origin point"
      when 4
        view.tooltip = "Click end X axis orientation"
      when 5
        view.tooltip = "Click end Y axis orientation"
      end

    end

    # When the user presses the left mouse button.
    def onLButtonDown(flags, x, y, view)
      # When the user clicks the first time, we switch to getting the
      # second point.  When they click a second time we create the line

      #puts "State: " + @state.to_s
      @ip1.pick view, x, y
      #if( @ip1.valid? )
      case @state
      when 0
        Sketchup::set_status_text("Origin", SB_PROMPT)
        @xdown = x
        @ydown = y
        view.tooltip = @ip1.tooltip if( @ip1.valid? )
        $align_points.push @ip1.position
        @state = 1
        view.invalidate
      when 1
        Sketchup::set_status_text("X Axis", SB_PROMPT)
        @xdown = x
        @ydown = y
        @ip2.pick view, x, y, @ip1
        view.tooltip = @ip2.tooltip if( @ip2.valid? )
        $align_points.push @ip2.position
        @state = 2
        view.invalidate
      when 2
        Sketchup::set_status_text("Y Axis", SB_PROMPT)
        @xdown = x
        @ydown = y
        @ip3.pick view, x, y, @ip1
        view.tooltip = @ip3.tooltip if( @ip3.valid? )
        $align_points.push @ip3.position
        @state = 3
        view.invalidate
      when 3
        Sketchup::set_status_text("End origin", SB_PROMPT)
        @xdown = x
        @ydown = y
        @ip4.pick view, x, y, @ip1
        view.tooltip = @ip4.tooltip if( @ip4.valid? )
        $align_points.push @ip4.position
        @state = 4
        view.invalidate
      when 4
        Sketchup::set_status_text("End X axis", SB_PROMPT)
        @xdown = x
        @ydown = y
        @ip5.pick view, x, y, @ip1
        view.tooltip = @ip5.tooltip if( @ip5.valid? )
        $align_points.push @ip5.position
        @state = 5
        view.invalidate
      when 5
        Sketchup::set_status_text("End Y axis", SB_PROMPT)
        @xdown = x
        @ydown = y
        @ip6.pick view, x, y, @ip1
        view.tooltip = @ip6.tooltip if( @ip6.valid? )
        $align_points.push @ip6.position
        @state = 6
        view.invalidate

        # Clear any inference lock
        view.lock_inference
      end #case
      #end
    end

    # user releases the left mouse button.
    def onLButtonUp(flags, x, y, view)
      self.reset(view)
      self.align_geometry(view) if @state == 6
    end

    # User presses a key on the keyboard.
    # We are checking it here to see if the user pressed the shift key
    # so that we can do inference locking
    def onKeyDown(key, repeat, flags, view)
      if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time = Time.now
        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
          # calling lock_inference with no arguments actually unlocks
          view.lock_inference
        elsif( @state == 0 && @ip1.valid? )
          view.lock_inference @ip1
        elsif( @state == 1 && @ip2.valid? )
          view.lock_inference @ip2
        elsif( @state == 2 && @ip3.valid? )
          view.lock_inference @ip3
        elsif( @state == 3 && @ip4.valid? )
          view.lock_inference @ip4
        elsif( @state == 4 && @ip5.valid? )
          view.lock_inference @ip5
        elsif( @state == 5 && @ip6.valid? )
          view.lock_inference @ip6
        end
      end
    end

    # User releases the key
    # We use this to unlock the interence
    # If the user holds down the shift key for more than 1/2 second, then we
    # unlock the inference on the release.  Otherwise, the user presses shift
    # once to lock and a second time to unlock.
    def onKeyUp(key, repeat, flags, view)
      return if not @state == 5
      if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
      end
    end

    # onUserText is called when the user enters something into the VCB
    def onUserText(text, view)
      return if not @state == 5
    end

    # The draw method is called whenever the view is refreshed.  It lets the
    # tool draw any temporary geometry that it needs to.
    def draw(view)
      if( @ip1.valid? )
        if( @ip1.display? )
          @ip1.draw(view)
          @drawn = true
        end
      end

      if $align_points.length == 1
        view.draw_points $align_points, 8, 2, "black"
      end

      if $align_points.length == 2
        view.line_width = 4
        view.draw_points $align_points[0], 8, 2, "black"
        #view.draw_points $align_points[1], 12, 7, "red"
        view.drawing_color = "red"
        view.draw_line $align_points[0], $align_points[1]
      end

      if $align_points.length == 3
        view.line_width = 4
        view.draw_points $align_points[0], 8, 2, "black"
        #view.draw_points $align_points[1], 12, 7, "red"
        #view.draw_points $align_points[2], 12, 7, "green"
        view.drawing_color = "red"
        view.draw_line $align_points[0], $align_points[1]
        view.drawing_color = "green"
        view.draw_line $align_points[0], $align_points[2]
        # Draw Z axis
        $v_start_x = $align_points[0].vector_to $align_points[1]
        $v_start_y = $align_points[0].vector_to $align_points[2]
        $v_start_z = $v_start_x.cross $v_start_y
        $v_start_z.length = ($v_start_x.length+$v_start_y.length)/2.0
        $p_start_z = $align_points[0].offset $v_start_z
        view.drawing_color = "blue"
        view.draw_line($align_points[0], $p_start_z)
      end

      if $align_points.length == 4
        view.line_width = 4
        view.draw_points $align_points[0], 8, 2, "black"
        #view.draw_points $align_points[1], 12, 7, "red"
        #view.draw_points $align_points[2], 12, 7, "green"
        view.draw_points $align_points[3], 8, 1, "black"
        view.drawing_color = "red"
        view.draw_line $align_points[0], $align_points[1]
        view.drawing_color = "green"
        view.draw_line $align_points[0], $align_points[2]
        view.drawing_color = "black"
        view.line_width = 1
        view.line_stipple = "-"
        view.draw_line $align_points[0], $align_points[3]
      end
      if $align_points.length == 5
        view.line_width = 4
        view.draw_points $align_points[0], 8, 2, "black"
        #view.draw_points $align_points[1], 12, 7, "red"
        #view.draw_points $align_points[2], 12, 7, "green"
        view.draw_points $align_points[3], 8, 1, "black"
        #view.draw_points $align_points[4], 12, 6, "red"
        view.drawing_color = "red"
        view.draw_line $align_points[0], $align_points[1]
        view.drawing_color = "green"
        view.draw_line $align_points[0], $align_points[2]
        view.drawing_color = "black"
        view.line_width = 1
        view.line_stipple = "-"
        view.draw_line $align_points[0], $align_points[3]
        view.line_stipple = "_"
        view.line_width = 4
        view.drawing_color = "red"
        view.draw_line $align_points[3], $align_points[4]
      end
      if $align_points.length == 6
        view.line_width = 4
        view.draw_points $align_points[0], 8, 2, "black"
        #view.draw_points $align_points[1], 12, 7, "red"
        #view.draw_points $align_points[2], 12, 7, "green"
        view.draw_points $align_points[3], 8, 1, "black"
        #view.draw_points $align_points[4], 12, 6, "red"
        #view.draw_points $align_points[5], 12, 6, "green"
        view.drawing_color = "red"
        view.draw_line $align_points[0], $align_points[1]
        view.drawing_color = "green"
        view.draw_line $align_points[0], $align_points[2]
        view.drawing_color = "black"
        view.line_width = 1
        view.line_stipple = "-"
        view.draw_line $align_points[0], $align_points[3]
        #view.line_stipple = ""
        view.line_width = 4
        view.line_stipple = "-"
        view.drawing_color = "red"
        view.draw_line $align_points[3], $align_points[4]
        view.drawing_color = "green"
        view.draw_line $align_points[3], $align_points[5]
        # Draw Z axis
        $v_end_x = $align_points[3].vector_to $align_points[4]
        $v_end_y = $align_points[3].vector_to $align_points[5]
        $v_end_z = $v_end_x.cross $v_end_y
        $v_end_z.length = ($v_end_x.length+$v_end_y.length)/2.0
        $p_end_z = $align_points[3].offset $v_end_z
        view.line_stipple = "_"
        view.drawing_color = "blue"
        view.draw_line $align_points[0], $p_start_z
        view.draw_line $align_points[3], $p_end_z

      end

      view.lock_inference

    end

    # onCancel is called when the user hits the escape key
    def onCancel(flag, view)
      self.reset(view)
      self.align_geometry(view) if @state == 5
      if $align_points.length < 6
        # We got less than 6 points
        UI.messagebox($align_points.length.to_s + " points are not enough !")
        self.reset(view)
      end
    end


    # Internal methods that are used to support the other methods in this class.

    # Reset the tool back to its initial state
    def reset(view)
      # This variable keeps track of which point we are currently getting
      #@state = 0
      case @state
      when 0
        Sketchup::set_status_text("Start origin", SB_PROMPT)
      when 1
        Sketchup::set_status_text("Start X axis", SB_PROMPT)
      when 2
        Sketchup::set_status_text("Start Y axis", SB_PROMPT)
      when 3
        Sketchup::set_status_text("End origin", SB_PROMPT)
      when 4
        Sketchup::set_status_text("End X axis", SB_PROMPT)
      when 5
        Sketchup::set_status_text("End Y axis", SB_PROMPT)
      when 6
        Sketchup::set_status_text("", SB_PROMPT)
      end

      # clear the InputPoint
      @ip1.clear
      if( view )
        view.tooltip = nil
        view.invalidate if @drawn
      end
      @drawn = false
      @dragging = false
    end

    # Create new geometry when the user has selected enough points.
    def align_geometry(view)

      ents_array=[]

      if $align_points.length == 6
        self.reset(view)
        # Build array of entities
        @sel.each do |e|
          ents_array.push e
        end

        # Vectors
        ###$vxbase = Geom::Vector3d.new(1, 0, 0)
        ###$vybase = Geom::Vector3d.new(0, 1, 0)
        ###$vzbase = Geom::Vector3d.new(0, 0, 1)

        ###$v_origins = $align_points[0].vector_to $align_points[3]

        #$v_start_x = $align_points[0].vector_to $align_points[1]
        #$v_start_y = $align_points[0].vector_to $align_points[2]
        #$v_end_x = $align_points[3].vector_to $align_points[4]
        #$v_end_y = $align_points[3].vector_to $align_points[5]

        #$v_start_z = $v_start_x.cross $v_start_y
        #$v_start_z.length = ($v_start_x.length+$v_start_y.length)/2.0
        #$p_start_z = $align_points[0].offset $v_start_z

        #$v_end_z = $v_end_x.cross $v_end_y
        #$v_end_z.length = ($v_end_x.length+$v_end_y.length)/2.0
        #$p_end_z = $align_points[3].offset $v_end_z

        # Rotation angle around X Y Z
        ###x_start = $v_start_x.angle_between $vxbase
        ###x_end = $v_end_x.angle_between $vxbase
        ###x_angle = x_end - x_start

        ###x_angle = $v_start_y.angle_between $v_end_y
        ###y_angle = $v_start_z.angle_between $v_end_z
        ###z_angle = $v_start_x.angle_between $v_end_x

        $v_start_x2=$v_start_x
        py2=$align_points[2].project_to_line [$align_points[0],$align_points[1]]
        $v_start_y2 = py2.vector_to $align_points[2]
        $v_start_z2=$v_start_x2.cross $v_start_y2

        $v_end_x2=$v_end_x
        py3=$align_points[5].project_to_line [$align_points[3],$align_points[4]]
        $v_end_y2 = py3.vector_to $align_points[5]
        $v_end_z2=$v_end_x2.cross $v_end_y2

        # Draw axes
        #Sketchup.active_model.entities.add_line($align_points[0],$align_points[0].offset($v_start_x2))
        #Sketchup.active_model.entities.add_line($align_points[0],$align_points[0].offset($v_start_y2))
        #Sketchup.active_model.entities.add_line($align_points[0],$align_points[0].offset($v_start_z2))
        #Sketchup.active_model.entities.add_line($align_points[3],$align_points[3].offset($v_end_x2))
        #Sketchup.active_model.entities.add_line($align_points[3],$align_points[3].offset($v_end_y2))
        #Sketchup.active_model.entities.add_line($align_points[3],$align_points[3].offset($v_end_z2))


        x_angle = $v_start_y2.angle_between $v_end_y2
        y_angle = $v_start_z2.angle_between $v_end_z2
        z_angle = $v_start_x2.angle_between $v_end_x2

        answer = 7
        if $comp == true
          case @sel[0].typename
          when "ComponentInstance"
            answer = UI.messagebox("Copy " + @sel[0].definition.name +  " component ?", MB_YESNO, "Option")
          when "Group"
            answer = UI.messagebox("Duplicate this group ?" , MB_YESNO, "Option")
          end
        end

        ###transfo1 = Geom::Transformation.axes $align_points[0], $v_start_x, $v_start_y, $v_start_z
        ###transfo2 = Geom::Transformation.axes $align_points[3], $v_end_x, $v_end_y, $v_end_z

        transfo1 = Geom::Transformation.axes $align_points[0], $v_start_x2, $v_start_y2, $v_start_z2
        transfo2 = Geom::Transformation.axes $align_points[3], $v_end_x2, $v_end_y2, $v_end_z2

        case answer
        when 7
          Sketchup.active_model.entities.transform_entities(transfo1.invert!, ents_array)
          Sketchup.active_model.entities.transform_entities(transfo2, ents_array)
        when 6
          case @sel[0].typename
          when "Group"
            new_group = @sel[0].copy
            Sketchup.active_model.entities.transform_entities(transfo1.invert!, new_group)
            Sketchup.active_model.entities.transform_entities(transfo2, new_group)
          when "ComponentInstance"
            # Get transformation of original
            org_trans = @sel[0].transformation
            new_comp = Sketchup.active_model.entities.add_instance @sel[0].definition, org_trans
            #new_comp = Sketchup.active_model.entities.add_instance @sel[0].definition, transfo1
            #new_comp.transformation = org_trans
            #new_comp.transformation = org_trans*transfo2
            Sketchup.active_model.entities.transform_entities(transfo1.invert!, new_comp)
            Sketchup.active_model.entities.transform_entities(transfo2, new_comp)
          end
        end

        # reset points array, old display axes, and exit
        $align_points=[]
        Sketchup.active_model.rendering_options["DisplayInstanceAxes"] = @mro_axes
        Sketchup.send_action "selectSelectionTool:"
        Sketchup.active_model.commit_operation
      end
    end

  end # class AlignTool

  #---------------------------------------
  # Check if there is something to align
  #---------------------------------------
  def Align.something_selected
    ss = Sketchup.active_model.selection
    if ss.empty?
      return nil
    else
      return true
    end
  end

  #---------------------------------------
  # Check if selection is a group/component
  #---------------------------------------
  def Align.component_selected
    ss = Sketchup.active_model.selection
    if ss.empty? or ss.length > 1
      r = false
    elsif ss.length == 1 and (ss[0].typename == "Group" or ss[0].typename == "ComponentInstance")
      r = true
    end
    r
  end

  #---------------------------------------
  # Select the Align tool
  #---------------------------------------
  def Align.tool
    $comp = false
    Sketchup.active_model.select_tool AlignTool.new
  end
  def Align.toolc
    $comp = true
    Sketchup.active_model.select_tool AlignTool.new
  end
end # module Align

#-----------------------------------------------------------------------------
# Shortcut for selecting the new tool within the UI
#-----------------------------------------------------------------------------
def align_tool
  Sketchup.active_model.select_tool AlignTool.new
end

if not $align_loaded
  main_menu = UI.menu("Tools").add_item("Align") { Align.tool }
  # Context menu handler
  UI.add_context_menu_handler do |menu|
    if( Align.something_selected )
      menu.add_item("Align") { Align.tool }
    end
    if( Align.component_selected )
      menu.add_item("Align Group/Component") { Align.toolc }
    end
  end
end
$align_loaded = true
