#----------------------------------------------------------------------------#
# Copyright 2010, Regular Polygon

# Permission to use this software without fee is hereby granted.
# Distribution or publication of this software, or substantial portions of
# the software, is subject to the expressed written consent of the author.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#----------------------------------------------------------------------------#
# Name        :   Torus 1.0
# Filename    :   rp_torus_v10.rb
# Version     :   1.0.1
# SU Version  :   6.0+
# Type        :   Dialog Box
# Description :   Creates a torus component. 
# Menu Item   :   Draw -> Torus
# Context Menu:   N/A
# Usage       :   Fill in the dialog box.
# Date        :   7/21/2010
# Author      :   Doug Herrmann
# Email       :   regular.polygon@gmail.com
# Website     :   http://regularpolygon.blogspot.com/
#----------------------------------------------------------------------------#


require 'sketchup.rb'


module RP

#----------------------------------------------------------------------------#
class Torus

	# Default dialog box parameters
	@@a = (Sketchup.read_default 'RP_torus', 'a', '12').to_f.inch       # inner radius
	@@b = (Sketchup.read_default 'RP_torus', 'b', '24').to_f.inch       # outer radius
	@@lat = (Sketchup.read_default 'RP_torus', 'lat', '36').to_i        # number of latitude lines
	@@long = (Sketchup.read_default 'RP_torus', 'long', '48').to_i      # number of longitude lines 
	@@twists = (Sketchup.read_default 'RP_torus', 'twists', '0').to_i   # number of edges it twists


	def initialize(a, b, lat, long, twists = 0)
		@r1, @r2, = (b - a)/2.0, (a + b)/2.0
		@lat, @long, @twists = lat, long, twists

		# make geomentry creation undoable
		Sketchup.active_model.start_operation "Definition"  
			 create_definition
		Sketchup.active_model.commit_operation 
	end


	# This method handles the Draw -> Torus menu item.
	def self.dialog

		# Show dialog to get user input for Torus paramters
		prompts = [
			'Inner Radius: ', 
			'Outer Radius: ', 
			'Latitude Lines: ', 
			'Longitude Lines: ', 
			'Twists: '
		]
		defaults = [@@a, @@b, @@lat, @@long, @@twists]
		results = nil

		# Keep prompting user until input is valid.
		begin
			results = UI.inputbox prompts, defaults, 'Torus Parameters'
			return unless results  # user cancelled the operation
			a, b, lat, long, twists = results
			raise "Invalid inner radius: #{ a }"  if ( a <= 0 )
			raise "The outer radius (#{ b }) must greater than the inner radius (#{ a })."  if ( b <= a )
			raise 'Enter at least 3 lines of latitude.'  if ( lat < 3 )
			raise 'Enter at least 3 lines of longitude.'  if ( long < 3 )
			raise 'The number of twists must be greater than or equal to zero.'  if ( twists < 0 )
		rescue
			UI.messagebox $!.message
			retry
		end
		
		# Create the Torus component and place it in the model.
		self.new(*results).place_component

		# Remember parameters
		@@a, @@b, @@lat, @@long, @@twists = results

		# Remember parameters between sessions
		Sketchup.write_default 'RP_torus', 'a', @@a.inspect
		Sketchup.write_default 'RP_torus', 'b', @@b.inspect
		Sketchup.write_default 'RP_torus', 'lat', @@lat.to_s
		Sketchup.write_default 'RP_torus', 'long', @@long.to_s
		Sketchup.write_default 'RP_torus', 'twists', @@twists.to_s
	end


	# Place an instance in the model using the component placement tool.
	def place_component
		Sketchup.active_model.place_component @definition 
	end

	
	private  #-----------------------------------------------------------------#


	# Create a Torus component definition
	def create_definition
		@definition = Sketchup.active_model.definitions.add 'Torus'
		entities = @definition.entities
		entities.fill_from_mesh create_mesh, true, 12
		@definition.insertion_point = Geom::Point3d.new(0, 0, 0)

		return  if ( (@long > 12) && (@lat > 12) && @twists.zero? )

		# Decide which edges to draw
		entities.each { |e|
			next  unless ( e.kind_of?(Sketchup::Edge) )

			# check if it's an internal edge
			if ( (@twists > 0) && ((e.start.edges.length == 4) || (e.end.edges.length == 4)) )
			e.smooth = true	
e.soft = true

                     # hide the profile
				next
			end

			# It is either a longitude or a latitude edge
			p1, v = e.line
			azimuth = Geom::Vector3d.new(-p1.y, p1.x, 0)

			if ( v.perpendicular? azimuth )
				# longitude edge
				if ( @long <= 12 )
					e.smooth = true
 e.soft = true         # draw it
				else
					e.smooth = true
e.soft = true                   # hide the profile
				end
			else
				# latitude edge
				if ( (@lat <= 12) || (@twists > 0) )
					e.smooth = false
e.soft = false        # draw it
				else
					e.smooth = false  
e.soft = false                 # hide the profile
				end
			end
		}
	end


	def create_mesh
		numpts = (@lat + 1) * (@long + 1)           # number of points in mesh
		numpts += (@lat * @long)  if ( @twists > 0 )
		numpoly = @lat * @long                      # number of polygons in mesh
		numpoly *= 4  if ( @twists > 0 )
		mesh = Geom::PolygonMesh.new(numpts, numpoly) 

		d_phi = 2 * Math::PI / @lat
		d_theta = 2 * Math::PI / @long
		d_twist = 2 * @twists * Math::PI / (@lat * @long)

		# Add points to mesh
		phi = 0 
		index_array = []
		for i in 0..@lat
			theta = 0
			twist = 0
			indices = []

			for j in 0..@long
				x = (@r2 + @r1 * Math.cos(phi + twist)) * Math.cos(theta)
				y = (@r2 + @r1 * Math.cos(phi + twist)) * Math.sin(theta)
				z = @r1 * Math.sin(phi + twist)
				indices.push( mesh.add_point([x, y, z]) )
				theta += d_theta
				twist += d_twist
			end

			index_array.push indices
			phi += d_phi
		end

		# Add center points of each square facet
		if ( @twists > 0 )
			center_index_array = []
			for i in 0...@lat
				i1 = index_array[i]
				i2 = index_array[(i + 1)]
				indices = []
				for j in 0...@long
					center = [0, 0, 0]
					center = Geom::Point3d.linear_combination 1, center, 0.25, mesh.point_at(i1[j])
					center = Geom::Point3d.linear_combination 1, center, 0.25, mesh.point_at(i1[j+1])
					center = Geom::Point3d.linear_combination 1, center, 0.25, mesh.point_at(i2[j+1])
					center = Geom::Point3d.linear_combination 1, center, 0.25, mesh.point_at(i2[j])
					indices.push( mesh.add_point(center) )
				end
				center_index_array.push indices
			end
		end

		# Add rows of polygons
		for i in 0...@lat do 
			i1 = index_array[i]
			i2 = index_array[(i + 1)]
			for j in 0...@long
				if @twists.zero? 
					mesh.add_polygon i1[j], i1[j+1], i2[j+1], i2[j]
				else
					center = center_index_array[i][j]
					mesh.add_polygon i1[j],   i1[j+1], center
					mesh.add_polygon i1[j+1], i2[j+1], center
					mesh.add_polygon i2[j+1], i2[j],   center
					mesh.add_polygon i2[j],   i1[j],   center
				end
			end
		end

		return mesh
	end
	
end  # class Torus

end  # module RP


# Menu
#----------------------------------------------------------------------------#
unless file_loaded? File.basename(__FILE__) 
	add_separator_to_menu('Draw')
	UI.menu('Draw').add_item('Torus') { RP::Torus.dialog } 
end

file_loaded File.basename(__FILE__)